

import json
import time
import logging
import numpy as np

from .primitive import LoopPrimitive
from pypot.utils.interpolation import KDTreeDict
logger = logging.getLogger(__name__)


class Move(object):

    """ Simple class used to represent a movement.

    This class simply wraps a sequence of positions of specified motors. The sequence must be recorded at a predefined frequency. This move can be recorded through the :class:`~pypot.primitive.move.MoveRecorder` class and played thanks to a :class:`~pypot.primitive.move.MovePlayer`.

    """

    def __init__(self, freq):
        self._framerate = freq
        self._timed_positions = KDTreeDict()

    def __repr__(self):
        return '<Move framerate={} #keyframes={}>'.format(self.framerate,
                                                          len(self.positions()))

    def __getitem__(self, i):
        return list(self._timed_positions.items())[i]

    @property
    def framerate(self):
        return self._framerate

    def add_position(self, pos, time):
        """ Add a new position to the movement sequence.

        Each position is typically stored as a dict of (time, (motor_name,motor_position)).
        """
        self._timed_positions[time] = pos

    def iterpositions(self):
        """ Returns an iterator on the stored positions. """
        return self._timed_positions.items()

    def positions(self):
        """ Returns a copy of the stored positions. """
        return self._timed_positions

    def plot(self, ax):
        pos = self.positions()

        if not pos:
            return

        motors = list(pos[0].keys())

        n = len(pos)
        t = np.linspace(0, n / self.framerate, n)
        pos = self.positions()

        p = {}
        for name in motors:
            p[name] = []

        for tt in t:
            for k, v in pos[float(tt)].items():
                p[k].append(v[0])

        for name in motors:
            ax.plot(t, p[name])

        ax.legend(motors)
        ax.set_xlabel('Time (in s)')
        ax.set_ylabel('Position (in degree)')

    def save(self, file):
        """ Saves the :class:`~pypot.primitive.move.Move` to a json file.

        .. note:: The format used to store the :class:`~pypot.primitive.move.Move` is extremely verbose and should be obviously optimized for long moves.
        """
        d = {
            'framerate': self.framerate,
            'positions': self._timed_positions,
        }
        json.dump(d, file, indent=2)

    @classmethod
    def create(cls, d):
        """ Create a :class:`~pypot.primitive.move.Move` from a dictionary. """
        move = cls(d['framerate'])
        move._timed_positions.update(d['positions'])
        return move

    @classmethod
    def load(cls, file):
        """ Loads a :class:`~pypot.primitive.move.Move` from a json file. """
        d = json.load(file)
        return cls.create(d)

    @classmethod
    def loads(cls, str):
        """ Loads a :class:`~pypot.primitive.move.Move` from a json string. """
        d = json.loads(str)
        return cls.create(d)


class MoveRecorder(LoopPrimitive):

    """ Primitive used to record a :class:`~pypot.primitive.move.Move`.

    The recording can be :meth:`~pypot.primitive.primitive.Primitive.start` and :meth:`~pypot.primitive.primitive.Primitive.stop` by using the :class:`~pypot.primitive.primitive.LoopPrimitive` methods.

    .. note:: Re-starting the recording will create a new :class:`~pypot.primitive.move.Move` losing all the previously stored data.

    """

    def __init__(self, robot, freq, tracked_motors):
        LoopPrimitive.__init__(self, robot, freq)
        self.freq = freq
        self.tracked_motors = list(map(self.get_mockup_motor, tracked_motors))

    def setup(self):
        self._move = Move(self.freq)

    def update(self):
        position = dict([(m.name, (m.present_position, m.present_speed))
                         for m in self.tracked_motors])
        self._move.add_position(position, self.elapsed_time)

    @property
    def move(self):
        """ Returns the currently recorded :class:`~pypot.primitive.move.Move`. """
        return self._move

    def add_tracked_motors(self, tracked_motors):
        """Add new motors to the recording"""
        new_mockup_motors = list(map(self.get_mockup_motor, tracked_motors))
        self.tracked_motors = list(set(self.tracked_motors + new_mockup_motors))


class MovePlayer(LoopPrimitive):

    """ Primitive used to play a :class:`~pypot.primitive.move.Move`.

    The playing can be :meth:`~pypot.primitive.primitive.Primitive.start` and :meth:`~pypot.primitive.primitive.Primitive.stop` by using the :class:`~pypot.primitive.primitive.LoopPrimitive` methods.

    .. warning:: the primitive is run automatically the same framerate than the move record.
        The play_speed attribute change only time lockup/interpolation
    """

    def __init__(self, robot, move=None, play_speed=1.0, move_filename=None, start_max_speed=50, **kwargs):
        self.move = move
        self.backwards = False
        if move_filename is not None:
            with open(move_filename, 'r') as f:
                self.move = Move.load(f)
        self.play_speed = play_speed if play_speed != 0 and isinstance(play_speed, float) else 1.0
        framerate = self.move.framerate if self.move is not None else 50.0
        self.start_max_speed = start_max_speed if start_max_speed != 0 else np.inf
        for key, value in kwargs.items():
            setattr(self, key, value)
        LoopPrimitive.__init__(self, robot, framerate)

    def setup(self):
        if self.move is None:
            raise AttributeError("Attribute move is not defined")
        self.period = 1.0 / self.move.framerate
        self.positions = self.move.positions()
        self.__duration = self.duration()
        if self.play_speed < 0:
            self.play_speed = - self.play_speed
            self.backwards = not self.backwards
        if self.play_speed == 0:
            self.play_speed = 1.0

        # Quick fix for limiting too fast movements at the play start
        max_goto_time = 0
        if self.backwards:
            position = self.positions[self.__duration]
        else:
            position = self.positions[0]
        for motor, value in position.items():
            motor = getattr(self.robot, motor)
            motor.compliant = False
            delta_angle = abs(motor.present_position - value[0])
            if delta_angle > 5:
                goto_time = delta_angle / self.start_max_speed
                motor.goto_position(value[0], goto_time)
                max_goto_time = goto_time if goto_time > max_goto_time else max_goto_time

        time.sleep(max_goto_time)

    def update(self):
        if self.elapsed_time < self.__duration:
            if self.backwards:
                position = self.positions[(self.__duration - self.elapsed_time) * self.play_speed]
            else:
                position = self.positions[self.elapsed_time * self.play_speed]

            for motor, value in position.items():
                # TODO: Ask pierre if its not a fgi to turn off the compliance
                getattr(self.robot, motor).compliant = False
                getattr(self.robot, motor).goal_position = value[0]
        else:
            self.stop()

    def duration(self):

        if self.move is not None:
            return (len(self.move.positions()) / self.move.framerate) / self.play_speed
        else:
            return 1.0



import json
import socket
import errno
import numpy
import logging

from tornado.ioloop import IOLoop
from tornado.web import RequestHandler
from tornado.web import Application

from .server import AbstractServer

logger = logging.getLogger(__name__)


class MyJSONEncoder(json.JSONEncoder):

    """ JSONEncoder which tries to call a json property before using the enconding default function. """

    def default(self, obj):
        if hasattr(obj, 'json'):
            return obj.json

        if isinstance(obj, numpy.ndarray):
            return list(obj)

        if isinstance(obj, numpy.integer):
            return int(obj)

        return json.JSONEncoder.default(self, obj)


class PoppyRequestHandler(RequestHandler):
    """Custom request handler.

    Automatically sets CORS and cache headers, and manages
    every OPTIONS request."""

    def set_default_headers(self):
        self.set_header('Cache-control', 'no-store')
        self.set_header('Access-Control-Allow-Origin', '*')
        self.set_header('Access-Control-Allow-Headers', 'Origin, Accept, Content-Type, X-Requested-With, X-CSRF-Token')
        self.set_header('Access-Control-Allow-Methods', 'GET, POST, PUT, DELETE, OPTIONS')

    def options(self, *args, **kwargs):
        self.set_status(204)

    def write_json(self, obj):
        self.write(json.dumps(obj, cls=MyJSONEncoder))


class IndexHandler(PoppyRequestHandler):
    def get(self, *args):
        out = {
            'motors': [],
            'primitives': []
        }
        for m in self.restful_robot.get_motors_list('motors'):
            motor = {}
            for r in self.restful_robot.get_motor_registers_list(m):
                try:
                    motor[r] = self.restful_robot.get_motor_register_value(m, r)
                except AttributeError:
                    pass
            out['motors'].append(motor)

        running_primitives = self.restful_robot.get_running_primitives_list()
        for prim in self.restful_robot.get_primitives_list():
            primitve = {
                'primitive': prim,
                'running': prim in running_primitives,
                'properties': [],
                # XXX pas de self en param ?
                'methods': self.restful_robot.get_primitive_methods_list(prim)
            }
            for prop in self.restful_robot.get_primitive_properties_list(prim):
                primitve['properties'].append({
                    'property': prop,
                    'value': self.restful_robot.get_primitive_property(prim, prop)
                })
            out['primitives'].append(primitve)

        self.write_json(out)


class MotorsListHandler(PoppyRequestHandler):
    def get(self, alias='motors'):
        self.write_json({
            'alias': self.restful_robot.get_motors_list(alias)
        })


class SensorsListHandler(PoppyRequestHandler):
    def get(self):
        self.write_json({
            'sensors': self.restful_robot.get_sensors_list()
        })


class MotorsAliasesListHandler(PoppyRequestHandler):
    def get(self):
        self.write_json({
            'alias': self.restful_robot.get_motors_alias()
        })


class MotorRegistersListHandler(PoppyRequestHandler):
    def get(self, motor_name):
        self.write_json({
            'registers': self.restful_robot.get_motor_registers_list(motor_name)
        })


class MotorRegisterHandler(PoppyRequestHandler):
    def get(self, motor_name, register_name):
        self.write_json({
            register_name: self.restful_robot.get_motor_register_value(motor_name, register_name)
        })


class UpdateMotorRegisterHandler(PoppyRequestHandler):
    def post(self, motor_name, register_name):
        data = json.loads(self.request.body.decode())
        self.restful_robot.set_motor_register_value(motor_name, register_name, data)
        self.write_json({})


class PrimitivesListHandler(PoppyRequestHandler):
    def get(self):
        self.write_json({
            'primitives': self.restful_robot.get_primitives_list()
        })


class RunningPrimitivesListHandler(PoppyRequestHandler):
    def get(self):
        self.write_json({
            'running_primitives': self.restful_robot.get_running_primitives_list()
        })


class StartPrimitiveHandler(PoppyRequestHandler):
    def get(self, primitive_name):
        self.restful_robot.start_primitive(primitive_name)


class StopPrimitiveHandler(PoppyRequestHandler):
    def get(self, primitive_name):
        self.restful_robot.stop_primitive(primitive_name)


class PausePrimitiveHandler(PoppyRequestHandler):
    def get(self, primitive_name):
        self.restful_robot.pause_primitive(primitive_name)


class ResumePrimitiveHandler(PoppyRequestHandler):
    def get(self, primitive_name):
        self.restful_robot.resume_primitive(primitive_name)


class PrimitivePropertiesListHandler(PoppyRequestHandler):
    def get(self, primitive_name):
        self.write_json({
            'property': self.restful_robot.get_primitive_properties_list(primitive_name)
        })


class PrimitivePropertyHandler(PoppyRequestHandler):
    def get(self, primitive_name, prop):
        response = self.restful_robot.get_primitive_property(primitive_name, prop)
        self.write_json({
            '{}.{}'.format(primitive_name, prop): response
        })


class SetPrimitivePropertyHandler(PoppyRequestHandler):
    def post(self, primitive_name, prop):
        data = json.loads(self.request.body.decode())
        self.restful_robot.set_primitive_property(primitive_name, prop, data)
        self.set_status(204)


class ListPrimitiveMethodsHandler(PoppyRequestHandler):
    def get(self, primitive_name):
        self.write_json({
            'methods': self.restful_robot.get_primitive_methods_list(primitive_name)
        })


class CallPrimitiveMethodHandler(PoppyRequestHandler):
    def post(self, primitive_name, method_name):
        data = json.loads(self.request.body.decode())
        response = self.restful_robot.call_primitive_method(primitive_name, method_name, data)
        self.write_json({
            '{}:{}'.format(primitive_name, method_name): response
        })


class MotorsRegistersHandler(PoppyRequestHandler):
    def get(self, register_name):
        motors_list = self.restful_robot.get_motors_list('motors')
        registers_motors = {}

        for motor_name in motors_list:
            registers_motors[motor_name] = {
                register_name: self.restful_robot.get_motor_register_value(
                    motor_name, register_name)
            }

        self.write_json(registers_motors)

class PathsUrl(PoppyRequestHandler):
    def get(self):
        out='<b>All url paths available:</b><br>'
        get='<br><b>Get method url:</b><br>'
        post='<br><b>Post method url:</b><br>'
        for url in url_paths:
            tmp=url[0]
            tmp=tmp.replace('\\','')
            tmp=tmp.replace('?','')
            tmp=tmp.replace('<','&lt')
            tmp=tmp.replace('>','&gt')
            tmp=tmp.replace('(P','')
            tmp=tmp.replace('[a-zA-Z0-9_]+)','')
            if 'value' in tmp:
                post+=tmp+'<br>'
            else:
                get+=tmp+'<br>'
        out+=get
        out+=post
        self.write(out)

url_paths=[
    (r'/', PathsUrl),
    (r'/robot\.json', IndexHandler),
    (r'/motor/alias/list\.json', MotorsAliasesListHandler),
    (r'/motor/(?P<alias>[a-zA-Z0-9_]+)/?list\.json', MotorsListHandler),
    (r'/sensor/list\.json', SensorsListHandler),
    (r'/motor/(?P<motor_name>[a-zA-Z0-9_]+)/register/list\.json', MotorRegistersListHandler),
    (r'/sensor/(?P<motor_name>[a-zA-Z0-9_]+)/register/list\.json', MotorRegistersListHandler),
    (r'/motor/(?P<motor_name>[a-zA-Z0-9_]+)/register/(?P<register_name>[a-zA-Z0-9_]+)/list\.json', MotorRegisterHandler),
    (r'/sensor/(?P<motor_name>[a-zA-Z0-9_]+)/register/(?P<register_name>[a-zA-Z0-9_]+)/list\.json', MotorRegisterHandler),
    (r'/motor/(?P<motor_name>[a-zA-Z0-9_]+)/register/(?P<register_name>[a-zA-Z0-9_]+)', MotorRegisterHandler),
    (r'/sensor/(?P<motor_name>[a-zA-Z0-9_]+)/register/(?P<register_name>[a-zA-Z0-9_]+)', MotorRegisterHandler),
    (r'/motor/(?P<motor_name>[a-zA-Z0-9_]+)/register/(?P<register_name>[a-zA-Z0-9_]+)/value\.json', UpdateMotorRegisterHandler),
    (r'/sensor/(?P<motor_name>[a-zA-Z0-9_]+)/register/(?P<register_name>[a-zA-Z0-9_]+)/value\.json', UpdateMotorRegisterHandler),
    (r'/primitive/list\.json', PrimitivesListHandler),
    (r'/primitive/running/list\.json', RunningPrimitivesListHandler),
    (r'/primitive/(?P<primitive_name>[a-zA-Z0-9_]+)/start\.json', StartPrimitiveHandler),
    (r'/primitive/(?P<primitive_name>[a-zA-Z0-9_]+)/stop\.json', StopPrimitiveHandler),
    (r'/primitive/(?P<primitive_name>[a-zA-Z0-9_]+)/pause\.json', PausePrimitiveHandler),
    (r'/primitive/(?P<primitive_name>[a-zA-Z0-9_]+)/resume\.json', ResumePrimitiveHandler),
    (r'/primitive/(?P<primitive_name>[a-zA-Z0-9_]+)/property/list\.json', PrimitivePropertiesListHandler),
    (r'/primitive/(?P<primitive_name>[a-zA-Z0-9_]+)/property/(?P<prop>[a-zA-Z0-9_]+)', PrimitivePropertyHandler),
    (r'/primitive/(?P<primitive_name>[a-zA-Z0-9_]+)/property/(?P<prop>[a-zA-Z0-9_]+)/value\.json', SetPrimitivePropertyHandler),
    (r'/primitive/(?P<primitive_name>[a-zA-Z0-9_]+)/method/list\.json', ListPrimitiveMethodsHandler),
    (r'/primitive/(?P<primitive_name>[a-zA-Z0-9_]+)/method/(?P<method_name>[a-zA-Z0-9_]+)/args\.json', CallPrimitiveMethodHandler),
    (r'/motors/register/(?P<register_name>[a-zA-Z0-9_]+)', MotorsRegistersHandler),
    ]


class HTTPRobotServer(AbstractServer):
    """Refer to the REST API for an exhaustive list of the possible routes."""

    def __init__(self, robot, host='0.0.0.0', port='8080', cross_domain_origin='*', **kwargs):
        AbstractServer.__init__(self, robot, host, port)

    def make_app(self):
        PoppyRequestHandler.restful_robot = self.restful_robot
        return Application(url_paths)

    def run(self, **kwargs):
        """ Start the tornado server, run forever"""

        try:
            loop = IOLoop()
            app = self.make_app()
            app.listen(self.port)
            loop.start()

        except socket.error as serr:
            # Re raise the socket error if not "[Errno 98] Address already in use"
            if serr.errno != errno.EADDRINUSE:
                raise serr
            else:
                logger.warning('The webserver port {} is already used. May be the HttpRobotServer is already running or another software is using this port.'.format(self.port))



#!/usr/bin/env python

from io import open
import re
import os
import sys

from setuptools import setup, find_packages


def version():
    with open('pypot/_version.py') as f:
        return re.search(r"^__version__ = ['\"]([^'\"]*)['\"]", f.read()).group(1)

install_requires = ['numpy',
                    'pyserial>2.6',
                    'tornado',
                    'scipy',
                    'ikpy==3.0.1',
                    'bottle',
                    'requests',
                    'opencv-contrib-python',
                    'wget',
                    ]

if sys.version_info < (3, 5):
    print("python version < 3.5 is not supported")
    sys.exit(1)

def package_files(directory):
    paths = []
    for (path, directories, filenames) in os.walk(directory):
        for filename in filenames:
            full_path = os.path.join(path, filename)
            paths.append((path, [full_path]))
    return paths


setup(name='pypot',
      version=version(),
      packages=find_packages(),

      install_requires=install_requires,

      extras_require={
          'doc': ['sphinx', 'sphinxjp.themes.basicstrap', 'sphinx-bootstrap-theme'],
          'zmq-server': ['zmq'],
          'remote-robot': ['zerorpc'],
          'camera': ['hampy', 'zmq'],  # Extras require: opencv (not a PyPi packet)
          'tests': ['requests', 'websocket-client', 'poppy-ergo-jr'],
      },

      entry_points={
          'console_scripts': [
              'dxl-config = pypot.tools.dxlconfig:main',
              'poppy-services=pypot.creatures.services_launcher:main',
              'poppy-configure=pypot.creatures.configure_utility:main',
          ],
      },

      include_package_data=True,
      exclude_package_data={'': ['.gitignore']},

      zip_safe=False,

      author='See https://github.com/poppy-project/pypot/graphs/contributors',
      author_email='dev@poppy-station.org',
      description='Python 3 Library for Robot Control',
      long_description=open('README.md', encoding='utf-8').read(),
      url='https://github.com/poppy-project/pypot',
      license='GNU GENERAL PUBLIC LICENSE Version 3',

      classifiers=[
          "Programming Language :: Python :: 3",
          "Topic :: Scientific/Engineering", ],
      )



try:
    from .marker import MarkerDetector
except ImportError:
    pass

try:
    from .blob import BlobDetector
except ImportError:
    pass

try:
    from .face import FaceDetector
except ImportError:
    pass
